# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

import serial
import time
import re

from htf.powersupply.core import SetOutputException, SetVoltageException, SetCurrentException, TimeoutException


def _ord(c):
    """ Python 3 and 2 ord method """
    if not isinstance(c, int):
        c = ord(c)
    return c


class KORAD_KA3005P:
    """
    ``KORAD_KA3005P`` is a hardware abstraction to the KORAD KA3005P power supply unit using a UART to control it.

    ``KORAD_KA3005P`` uses ``DelegatorMixin`` to automatically set and read it's members ``voltage`` and ``current``.
    """

    def __init__(self,
                 port,
                 baudrate=9600,
                 parity=serial.PARITY_NONE,
                 stopbits=serial.STOPBITS_ONE,
                 timeout=0.5,
                 debug=False):
        """
        Args:
            port (str): the comport to be used, eg. "COM3", 3 or "/dev/ttyUSB0", etc.
            debug=False (bool): if set to ``True`` debugging is enabled.
        """
        super(KORAD_KA3005P, self).__init__()

        # open the serial port
        try:
            self.serial = serial.serial_for_url(
                port,
                baudrate=baudrate,
                parity=parity,
                stopbits=stopbits,
                timeout=timeout
            )

            # check if the device responds
            self.get_status()

            self.debug = debug

            self.id = self.query(b"*IDN?")
            if self.debug:
                print(self.__class__.__name__, "found", self.id)
            if re.match(r"^KORAD\s*KA3005P.*", self.id.decode()) is None:
                raise Exception("Found unsupported device", self.id)

        except Exception:
            if hasattr(self, "serial"):
                self.serial.close()
            raise

    # properties
    @property
    def voltage(self):
        return self.get_current_voltage()

    @voltage.setter
    def voltage(self, value):
        self.set_voltage(value)

    @property
    def current(self):
        return self.get_current_current()

    @current.setter
    def current(self, value):
        self.set_current(value)

    def get_id(self):
        """
        Return the KORAD id.

        Returns:
            str: the KORAD id.
        """
        return self.id

    getId = get_id

    def send_command(self, command):
        """
        Directly send a command to the power supply unit.

        Args:
            command (str): the command to be sent.
        """
        if not isinstance(command, bytes):  # Python 3 hack
            command = command.encode()
        self.serial.write(command)
        time.sleep(.05)

    sendCommand = send_command

    def query(self, query, length_of_answer=None, retries=3):
        """
        Send a command and wait for the answer.

        Args:
            query (str): the query to be sent.
            length_of_answer=None (int): the maximum number of bytes to be received.
            retries=3 (int): the number of retries.

        Returns:
            str: the answer.

        Raises:
            korad_ka3005p.TimeoutException: if a communication time-out occurred.
        """

        if not isinstance(query, bytes):  # Python 3 hack
            query = query.encode()

        while retries:
            self.serial.write(query)
            if length_of_answer is not None:
                answer = self.serial.read(length_of_answer)
            else:
                answer = self.serial.read(1000)
            if answer not in [None, ""]:
                return answer

            retries -= 1

        raise TimeoutException("Device does not respond!")

    def get_status(self):
        """
        Read device status and return a dictionary.

        Returns:
            dict: the status dictionary containing the keys "ch1 mode", "ch2 mode", "tracking mode", "beep",
             "lock" and "output".
        """
        status_string = self.query(b"STATUS?", 1)

        byte = _ord(status_string[0])

        status = {
            "ch1 mode": ["cc", "cv"][byte & 0x01],
            "ch2 mode": ["cc", "cv"][(byte & 0x02) >> 1],
            "tracking mode": ["independent", "tracking series", "tracking parallel"][(byte & 0x0c) >> 3],
            "beep": ["off", "on"][(byte & (0x01 << 4)) >> 4],
            "lock": ["off", "on"][(byte & (0x01 << 5)) >> 5],
            "output": ["off", "on"][(byte & (0x01 << 6)) >> 6]
        }

        # 0 CH1 0=CC mode, 1=CV mode
        # 1 CH2 0=CC mode, 1=CV mode
        # 2, 3 Tracking 00=Independent, 01=Tracking series,11=Tracking parallel
        # 4 Beep 0=Off, 1=On
        # 5 Lock 0=Lock, 1=Unlock
        # 6 Output 0=Off, 1=On
        # 7 N/A N/A

        return status

    getStatus = get_status

    def enable_output(self):
        """
        Enable output.

        Raises:
            SetOutputException: if output was not set.
        """
        self.send_command(b"OUT1")

        status = self.get_status()

        if not status["output"] == "on":
            raise SetOutputException("The output was not enabled!")

    enableOutput = enable_output

    def disable_output(self):
        """
        Disable output.
        """
        self.send_command(b"OUT0")

        status = self.get_status()

        if not status["output"] == "off":
            raise SetOutputException("The output was not enabled!")

    disableOutput = disable_output

    def set_voltage(self, voltage, channel=1):
        """
        Set the output voltage for a given channel.

        If ``voltage`` is None the output is disabled.
        If ``voltage`` is not None the output is enabled.
        If ``voltage`` changes the output is first disabled, changed, and then enabled again.

        Args:
            voltage (float or None): the voltage in units of volts as int(), float() or string, eg. 23, 23.0, "23.0"
            or even "23.0 V" work.
            channel (int): the channel to use. Shall always be 1.

        Raises:
            SetVoltageException: if voltage was not set.
        """
        if voltage is None:
            self.disable_output()
        else:
            voltage_string = ("%05.2f" % float(voltage)).encode()
            command = ("VSET%i:" % channel).encode() + voltage_string
            self.send_command(command)
            self.enable_output()

            query = ("VSET%i?" % channel).encode()
            voltage_setting = self.query(query, 5)

            if not voltage_setting[:5] == voltage_string[:5]:
                raise SetVoltageException("Setting voltage to " + voltage_string +
                                          " did not work. Setting in KORAD is " + voltage_setting)

    setVoltage = set_voltage

    def get_current_voltage(self, channel=1):
        """
        Read current voltage from device and return the result.

        Args:
            channel=1 (int): the channel to use. Shall always be 1.

        Returns:
            float: the current voltage.
        """
        query = ("VOUT%i?" % channel).encode()
        voltage_currently = self.query(query, 5)
        return float(voltage_currently)

    getCurrentVoltage = get_current_voltage

    def set_current(self, current, channel=1):
        """
        Set the output current for a given channel.

        Args:
            current (int, float or str): the current in units of amperes as int(), float() or string, eg. 0, 0.5, "0.5"
                or even "0.5 mA" work.
            channel (int): the channel to use. Shall always be 1.

        Raises:
            SetCurrentException: if current was not set.
        """
        current_string = ("%05.3f" % float(current)).encode()
        command = ("ISET%i:" % channel).encode() + current_string
        self.send_command(command)

        query = ("ISET%i?" % channel).encode()
        current_setting = self.query(query, 5)
        if not current_setting[:5] == current_string[:5]:
            raise SetCurrentException("Setting current to " + current_string +
                                      " did not work. Setting in KORAD is " + current_setting)

    setCurrent = set_current

    def get_current_current(self, channel=1):
        """
        Read the current current from device and return the result.

        Args:
            channel (int): the channel to use. Shall always be 1.

        Returns:
            float: the current current.
        """
        query = ("IOUT%i?" % channel).encode()
        current_currently = self.query(query, 5)
        return float(current_currently)

    getCurrentCurrent = get_current_current

    def close(self):
        """
        Close the serial member. This method is automatically called when :func:`__del__` is called.
        """
        if hasattr(self, "serial"):
            self.serial.close()

    def __del__(self):
        """
        Destructor of an ``KORAD_KA3005P`` instance.
        This method calls :func:`close()` to automatically close the serial instance.
        """
        self.close()
