# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

import serial
import time
import threading
from htf.powersupply.core import ChannelVoltage, ActualChannelVoltage, ChannelCurrent, ActualChannelCurrent
import logging
# logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class _HMPSeriesAbstraction:
    """
    An SCPI interface for all Rohde & Schwarz HMP power supplies.
    """
    NUMBER_OF_CHANNELS = None
    IDNS = []

    voltage1 = ChannelVoltage(1)
    voltage2 = ChannelVoltage(2)
    voltage3 = ChannelVoltage(3)
    voltage4 = ChannelVoltage(4)

    actual_voltage1 = ActualChannelVoltage(1)
    actual_voltage2 = ActualChannelVoltage(2)
    actual_voltage3 = ActualChannelVoltage(3)
    actual_voltage4 = ActualChannelVoltage(4)

    current1 = ChannelCurrent(1)
    current2 = ChannelCurrent(2)
    current3 = ChannelCurrent(3)
    current4 = ChannelCurrent(4)

    actual_current1 = ActualChannelCurrent(1)
    actual_current2 = ActualChannelCurrent(2)
    actual_current3 = ActualChannelCurrent(3)
    actual_current4 = ActualChannelCurrent(4)

    def __init__(self,
                 comport,
                 baudrate=9600,
                 parity=serial.PARITY_NONE,
                 stopbits=serial.STOPBITS_ONE,
                 timeout=1.0,
                 command_delay=0.05,
                 wait_after_set_voltage=False,  # fast changes in setting voltage is possible, but be aware, that the
                                                # device needs time for reaching setpoint level
                 wait_time=0.5):

        self._command_delay = command_delay
        self._wait_after_set_voltage = wait_after_set_voltage
        self._wait_time = wait_time
        self._serial = serial.serial_for_url(
            comport,
            baudrate=baudrate,
            parity=parity,
            stopbits=stopbits,
            timeout=timeout,
            write_timeout=timeout,
            rtscts=True)
        self._lock = threading.Lock()

        self._check_idn()

    def close(self):
        """
        Close serial connection.
        """
        if hasattr(self, "_serial"):
            self._serial.close()

    def __del__(self):
        self.close()

    def _check_idn(self):
        device_idn = self.get_idn()

        # loop over all possible IDNs and raise exception if no IDN matches

        for idn in self.IDNS:
            if device_idn.startswith(idn):
                return

        raise AssertionError(
            "Expected IDN to start with %s but found %s. Select the right powersupply abstractor." % (
                b", ".join(self.IDNS), device_idn))

    def _set_voltage_for_channel_generator(self, channel):
        def _set_voltage_for_channel(voltage):
            self.select_channel(channel)
            self.set_voltage(voltage, auto_enable=True)
        return _set_voltage_for_channel

    def _get_voltage_for_channel_generator(self, channel):
        def _get_voltage_for_channel():
            self.select_channel(channel)
            return self.get_voltage()
        return _get_voltage_for_channel

    def _get_actual_voltage_for_channel_generator(self, channel):
        def _get_actual_voltage_for_channel():
            self.select_channel(channel)
            return self.get_actual_voltage()
        return _get_actual_voltage_for_channel

    def _set_current_for_channel_generator(self, channel):
        def _set_current_for_channel(current):
            self.select_channel(channel)
            self.set_current(current)
        return _set_current_for_channel

    def _get_current_for_channel_generator(self, channel):
        def _get_current_for_channel():
            self.select_channel(channel)
            return self.get_current()
        return _get_current_for_channel

    def _get_actual_current_for_channel_generator(self, channel):
        def _get_actual_current_for_channel():
            self.select_channel(channel)
            return self.get_actual_current()
        return _get_actual_current_for_channel

    def _query(self, commands, response_length=0):
        with self._lock:
            self._serial.flushInput()

            for command in commands:
                command = b"\r\n" + command + b"\r\n"
                logger.debug("send: '%s'" % command)
                self._serial.write(command)
                time.sleep(self._command_delay)

            response = b""
            received = 0

            while received < response_length:
                b = self._serial.read(1)
                logger.debug("received '%s'" % b)
                if b in [b"", b"\n"]:
                    break
                response += b
                received += 1

        return response

    def get_idn(self):
        """
        Get the identifier.

        Returns:
            str: the identifier of the device
        """
        idn = self._query([b"*IDN?"], response_length=100)
        return idn

    def select_channel(self, channel):
        """
        Select a channel.

        Args:
            channel (int): the channel to be selected (``1 .. 4``)
        """
        assert 1 <= channel <= 4
        ret = self._query([b"INST OUT%i" % channel, b"INST?"], response_length=5)
        assert ret == b"OUTP%i" % channel

    def select_output_channel(self):
        """
        Select currently selected channel for general output.
        """
        self._query([b"OUTP:SEL ON"], response_length=0)

    def deselect_output_channel(self):
        """
        Deselect currently selected channel for general output.
        """
        self._query([b"OUTP:SEL OFF"], response_length=0)

    def enable_output(self):
        """
        Enable the output of the currently selected channel.
        """
        ret = self._query([b"OUTP ON", b"OUTP?"], response_length=1)
        assert ret == b"1"

    def disable_output(self):
        """
        Disable the output of the currently selected channel.
        """
        ret = self._query([b"OUTP OFF", b"OUTP?"], response_length=1)
        assert ret == b"0"

    def enable_general_output(self):
        """
        Enable the general output.
        """
        ret = self._query([b"OUTP:GEN ON", b"OUTP?"], response_length=1)
        assert ret == b"1"

    def disable_general_output(self):
        """
        Enable the general output.
        """
        ret = self._query([b"OUTP:GEN OFF", b"OUTP?"], response_length=1)
        assert ret == b"1"

    def store_state(self, location):
        """
        Save the current instrument state to the specified storage location.

        Args:
            location (int): the storage location to store the instrument state to.
        """
        assert 0 <= location < 10, "location must be between 0 and 9"
        self._query([b"*SAV %i" % location], response_length=0)

    def recall_state(self, location):
        """
        Recall the instrument state from the specified storage location.

        Args:
            location (int): the storage location to load the instrument state from.
        """
        assert 0 <= location < 10, "location must be between 0 and 9"
        ret = self._query([b"*RCL %i" % location], response_length=0)

    def set_voltage(self, voltage, auto_enable=False):
        """
        Set the voltage on the currently selected channel.

        Args:
            voltage (float): the voltage to be set.
            auto_enable=False (bool): if set to ``True`` the output is enabled automatically.
        """
        if voltage is not None:
            ret = self._query([b"VOLT %.3f" % voltage, b"VOLT?"], response_length=6)
            expected_voltage = b"%.3f" % voltage
            assert ret == expected_voltage
            if auto_enable is True:
                self.enable_output()
            if self._wait_after_set_voltage is True:
                logger.debug("wait after set voltage to %f V for %f s" % (voltage, self._wait_time))
                time.sleep(self._wait_time)
        else:
            if auto_enable is True:
                self.disable_output()

    def get_voltage(self):
        """
        Get the voltage of the currently selected channel.

        Returns:
            float: the voltage of the currently selected channel
        """
        ret = self._query([b"VOLT?"], response_length=6)
        return float(ret)

    def get_actual_voltage(self):
        """
        Get the measured voltage of the currently selected channel.

        Returns:
            float: the measured voltage of the currently selected channel
        """
        ret = self._query([b"MEAS:VOLT?"], response_length=6)
        return float(ret)

    def set_current(self, current, read_back=True):
        """
        Set the current of the currently selected channel

        Args:
            current (float): the current to be set
            read_back=True (bool): if True the set current is read back and compared to the expected current
        """
        current = round(current, 3)
        logger.debug("_set_current: %f", current)
        ret = self._query([b"CURR %.3f" % current, b"CURR?"], response_length=6)
        if read_back:
            if current < 10:
                expected_current = b"%.4f" % current  # 5 significant places
            else:
                expected_current = b"%.3f" % current  # 5 significant places
            logger.debug("expected_current %s" % expected_current)
            assert ret == expected_current, "read back current mismatch: expected '%s' but '%s' was recieved" % (expected_current, ret)

    def get_current(self):
        """
        Get the current of the currently selected channel.

        Returns:
            float: the current of the currently selected channel
        """
        ret = self._query([b"CURR?"], response_length=6)
        return float(ret)
    
    def get_actual_current(self):
        """
        Get the measured current of the currently selected channel.

        Returns:
            float: the measured current of the currently selected channel
        """
        ret = self._query([b"MEAS:CURR?"], response_length=6)
        return float(ret)


class HMP2020(_HMPSeriesAbstraction):
    """
    HMP2020: 1 × 0 V to 32 V/0 A to 10 A; 1 × 0 V to 32 V/0 A to 5 A (188 W)
    """
    NUMBER_OF_CHANNELS = 2
    IDNS = [b"HAMEG,HMP2020,", b"ROHDE&SCHWARZ,HMP2020,"]


class HMP2030(_HMPSeriesAbstraction):
    """
    HMP2030: 3 × 0 V to 32 V/0 A to 5 A (188 W)
    """
    NUMBER_OF_CHANNELS = 3
    IDNS = [b"HAMEG,HMP2030,", b"ROHDE&SCHWARZ,HMP2030,"]


class HMP4030(_HMPSeriesAbstraction):
    """
    HMP4030: 3 × 0 V to 32 V/0 A to 10 A (384 W)
    """
    NUMBER_OF_CHANNELS = 3
    IDNS = [b"HAMEG,HMP4030,", b"ROHDE&SCHWARZ,HMP4030,"]


class HMP4040(_HMPSeriesAbstraction):
    """
    HMP4040: 4 × 0 V to 32 V/0 A to 10 A (384 W)
    """
    NUMBER_OF_CHANNELS = 4
    IDNS = [b"HAMEG,HMP4040,", b"ROHDE&SCHWARZ,HMP4040,"]
